import re
import os
import glob
from typing import Tuple, List

import numpy as np
import cv2
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from .utils import binarize, read_and_undis

def unwrap_phase_single(wrapped_phase_low, wrapped_phase_high, ratio):
    # 相位展开
    unwrapped_phase = wrapped_phase_high + 2 * np.pi * np.round((wrapped_phase_low * ratio - wrapped_phase_high) / (2 * np.pi))
    return unwrapped_phase

def natural_sort_key(s):
    # 提取字符串中的数字，转换为整数并返回作为排序键
    return [int(text) if text.isdigit() else text for text in re.split(r'(\d+)', s)]

def sort_filenames(filenames):
    return sorted(filenames, key=natural_sort_key)

def read_image_data(image_folder: str, cam_id: int, mtx, dst, img_type: str = 'bmp', num_freq: int = 2, num_shift: int = 4) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    读取图像数据并返回不同频率和方向的图像数组。

    Args:
        image_folder (str): 图像文件夹路径
        img_type (str): 图像文件类型
        num_freq (int): 频率数量
        num_shift (int): 相移步数

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: 低频x, 低频y, 高频x, 高频y图像数组
    """
    filenames = glob.glob(os.path.join(image_folder, f'*{str(cam_id)}*_*.{img_type}'))
    #print('filenames = ',filenames)
    filenames = sort_filenames(filenames)
    # print(filenames)
    # import pdb; pdb.set_trace()
    assert len(filenames) >= 2*num_freq*num_shift, f'There should be 2*{num_freq}*{num_shift} images in the folder, but got {len(filenames)}'
    
    img_data = []
    for img_path in filenames[:2*num_freq*num_shift]:
        img = read_and_undis(img_path, mtx, dst)
        if img is None:
            raise ValueError(f"Failed to read image: {img_path}")
        img_data.append(img.astype(np.float64))
    
    img_data = np.array(img_data).transpose(1, 2, 0)
  
    output_img = []
    # import pdb; pdb.set_trace()
    for i in range(num_freq):
        n = i*2
        y = img_data[..., n*num_shift:(n+1)*num_shift]
        x = img_data[..., (n+1)*num_shift:(n+2)*num_shift]
        output_img.append(x)
        output_img.append(y)
    
    return output_img

def compute_phase(img_data: np.ndarray, mask: np.ndarray) -> np.ndarray:
    """
    计算相位信息。

    Args:
        img_data (np.ndarray): 输入图像数据
        mask (np.ndarray): 掩码数组

    Returns:
        np.ndarray: 计算得到的相位数组
    """
    I1, I2, I3, I4 = [np.ma.masked_where(mask == 0, img_data[:, :, i]) for i in range(4)]
    phase = np.arctan2(I3 - I1, I4 - I2) + np.pi
    return phase.filled(fill_value=np.nan)

def extract_phase_2(img_folder: str, cam_id: int, mask, mtx, dst, num_shift: int = 4) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:

    """
    提取相位信息。

    Args:
        img_folder (str): 图像文件夹路径
        num_freq (int): 频率数量
        num_shift (int): 相移数量

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: 低频x相位, 低频y相位, 高频x相位, 高频y相位
    """
    print(img_folder)
    high_freq_x, high_freq_y, low_freq_x, low_freq_y = read_image_data(img_folder, mtx, dst,num_freq=2, num_shift=num_shift)

    return (
        compute_phase(high_freq_x, mask),
        compute_phase(high_freq_y, mask),
        compute_phase(low_freq_x, mask),
        compute_phase(low_freq_y, mask)
    )

def extract_phase_3(img_folder: str, cam_id: int, mask, mtx, dst, num_shift: int = 4) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    high_freq_x, high_freq_y, mid_freq_x, mid_freq_y, low_freq_x, low_freq_y = read_image_data(img_folder, cam_id, mtx, dst, num_freq=3, num_shift=num_shift)

    return (
        compute_phase(high_freq_x, mask),
        compute_phase(high_freq_y, mask),
        compute_phase(mid_freq_x, mask),
        compute_phase(mid_freq_y, mask),
        compute_phase(low_freq_x, mask),
        compute_phase(low_freq_y, mask)
    )

def extract_phase(img_folder: str, cam_id: int, mask, mtx, dst, num_freq: int = 2, num_shift: int = 4) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    # mask_filename = os.path.join(img_folder, 'white.bmp')
    # mask_img = cv2.imread(mask_filename, cv2.IMREAD_GRAYSCALE)
    # if mask_img is None:
    #     raise ValueError(f"Failed to read mask image: {mask_filename}")
    # # TODO: 这里需要根据调制度调整阈值
    # mask = binarize(mask_img)
    
    if num_freq == 2:
        return extract_phase_2(img_folder, cam_id, mask, mtx, dst, num_shift)
    elif num_freq == 3:
        return extract_phase_3(img_folder, cam_id, mask, mtx, dst, num_shift)
    else:
        raise ValueError(f"num_freq should be 2 or 3, but got {num_freq}")

def unwrap_phase(phases, save_path, num_freq: int = 2, k: int = 32, debug: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    """
    解包裹相位。

    Args:
        phase (np.ndarray): 输入相位

    Returns:
        np.ndarray: 解包裹相位
    """
    if num_freq == 2:
        high_freq_x_phase, high_freq_y_phase, low_freq_x_phase, low_freq_y_phase = phases
        # 计算短周期的相位展开
        x_phase_unwrapped = unwrap_phase_single(low_freq_x_phase, high_freq_x_phase, k)
        y_phase_unwrapped = unwrap_phase_single(low_freq_y_phase, high_freq_y_phase, k)
    elif num_freq == 3:
        high_freq_x_phase, high_freq_y_phase, mid_freq_x_phase, mid_freq_y_phase, low_freq_x_phase, low_freq_y_phase = phases

        unwrapped_phase_mid_x = unwrap_phase_single(low_freq_x_phase, mid_freq_x_phase, 8)
        x_phase_unwrapped = unwrap_phase_single(unwrapped_phase_mid_x, high_freq_x_phase, 32/8)

        unwrapped_phase_mid_y = unwrap_phase_single(low_freq_y_phase, mid_freq_y_phase, 8)
        y_phase_unwrapped = unwrap_phase_single(unwrapped_phase_mid_y, high_freq_y_phase, 32/8)
      
    else:
        raise ValueError(f"num_freq should be 2 or 3, but got {num_freq}")

    if debug:

        plt.imshow(high_freq_y_phase)
        plt.title("high_freq_x_phase")
        #plt.savefig(os.path.join(save_path, "high_freq_x_phase.png"))
        plt.show()
        plt.imshow(high_freq_x_phase)
        plt.title("high_freq_y_phase")
        #plt.savefig(os.path.join(save_path, "high_freq_y_phase.png"))
        plt.show()

        x, y = np.meshgrid(np.arange(high_freq_x_phase.shape[1]), np.arange(high_freq_x_phase.shape[0]))

        fig, axes = plt.subplots(1, 2, figsize=(12, 6))

        # 第一个子图
        cax0 = axes[0].imshow(y_phase_unwrapped)
        axes[0].set_title('x_phase_unwrapped')
        axes[0].set_xlabel('X Axis')
        axes[0].set_ylabel('Y Axis')
        fig.colorbar(cax0, ax=axes[0])

        # 第二个子图
        cax1 = axes[1].imshow(x_phase_unwrapped)
        axes[1].set_title('y_phase_unwrapped')
        axes[1].set_xlabel('X Axis')
        axes[1].set_ylabel('Y Axis')
        fig.colorbar(cax0, ax=axes[1])

        # 调整子图之间的间距
        plt.tight_layout()
        #plt.savefig(os.path.join(save_path, "phase_unwrapped.png"))
        plt.show()

    return y_phase_unwrapped, x_phase_unwrapped